"""Renderers for django-secretary."""
import logging
import os
import subprocess  # nosec
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Union, cast

import weasyprint
from django.core.exceptions import ImproperlyConfigured
from django.template import loader
from django.template.backends.django import DjangoTemplates, Template as DjangoTemplate  # type: ignore[attr-defined]
from rest_framework.renderers import BaseRenderer

from .utils.fetchers import FileFetcher

if TYPE_CHECKING:
    from .models import Template as TemplateModel
    from .utils.templates import Template

_LOGGER = logging.getLogger(__name__)


class TemplateRendererMixin:
    """A mixin with utilities for template renderers.

    Class attributes:
        preferred_backend: Name of the preferred template backend.
        exception_template: A template for rendering an error.
    """

    preferred_backend: ClassVar[Optional[str]] = None
    exception_template: str = 'django_secretary/render_error.txt'

    def _get_backend(self) -> DjangoTemplates:
        """Return template backend to be used."""
        from .utils.templates import TEMPLATE_ENGINES
        if not TEMPLATE_ENGINES.all():
            raise ImproperlyConfigured("No secretary template backend is defined.")
        if self.preferred_backend in TEMPLATE_ENGINES:
            return cast(DjangoTemplates, TEMPLATE_ENGINES[cast(str, self.preferred_backend)])
        else:
            # XXX: Use only the first engine for now.
            return cast(DjangoTemplates, TEMPLATE_ENGINES.all()[0])

    def _get_template(self, renderer_context: Dict[str, Any]) -> Union['Template', DjangoTemplate]:
        """Return template to be rendered."""
        from .utils.templates import Template
        if renderer_context['response'].exception:
            return cast(DjangoTemplate, loader.get_template(self.exception_template))

        model = cast('TemplateModel', renderer_context['view'].get_object())
        return Template(model, self._get_backend())

    def _get_template_context(self, data: Any, renderer_context: Dict[str, Any]) -> Any:
        """Return a context data for the template."""
        if renderer_context['response'].exception:
            return {'errors': data}
        return data


class TextRenderer(TemplateRendererMixin, BaseRenderer):
    """Renderer which renders secretary template as a plain text.

    Attributes:
        exception_template: A template for rendering an error.
    """

    media_type = 'text/plain'
    format = 'txt'

    preferred_backend = 'text'

    def render(self, data: Any, media_type: str = None, renderer_context: Dict[str, Any] = None) -> str:
        """Render secretary template as a plain text."""
        renderer_context = renderer_context or {}
        return self._get_template(renderer_context).render(self._get_template_context(data, renderer_context))


class HtmlRenderer(TextRenderer):
    """Renderer which renders secretary template as HTML."""

    media_type = 'text/html'
    format = 'html'

    preferred_backend = 'html'
    exception_template = 'django_secretary/render_error.html'


class PdfRenderer(TemplateRendererMixin, BaseRenderer):
    """Renderer which renders secretary template as PDF.

    This rendered shouldn't be used with SignedPdfRenderer.
    """

    media_type = 'application/pdf'
    format = 'pdf'

    preferred_backend = 'html'

    def render(self, data: Any, media_type: str = None, renderer_context: Dict[str, Any] = None) -> bytes:
        """Render secretary template as PDF."""
        renderer_context = renderer_context or {}
        rendered_template = self._get_template(renderer_context).render(
            self._get_template_context(data, renderer_context))
        request = renderer_context['request']
        fetcher = FileFetcher(request.build_absolute_uri())
        pdf_renderer = weasyprint.HTML(string=rendered_template, url_fetcher=fetcher.fetch,
                                       base_url=request.build_absolute_uri())
        return cast(bytes, pdf_renderer.write_pdf())


class SignedPdfRenderer(PdfRenderer):
    """Renderer which renders secretary template as PDF and signs it.

    This rendered shouldn't be used with PdfRenderer.
    """

    def render(self, data: Any, media_type: str = None, renderer_context: Dict[str, Any] = None) -> bytes:
        """Render secretary template as PDF and sign it."""
        from .settings import SECRETARY_SETTINGS

        raw_pdf = super().render(data, media_type, renderer_context)
        template = self._get_template(renderer_context or {})
        signer = getattr(getattr(template, 'model', None), 'signer', None)

        if not signer:
            return raw_pdf

        if signer not in SECRETARY_SETTINGS.SIGNERS:
            _LOGGER.error("Settings for '%s' signer not found.", signer)
            return raw_pdf

        signer_setting = SECRETARY_SETTINGS.SIGNERS[signer]
        with TemporaryDirectory() as tmp_dir:
            with open(os.path.join(tmp_dir, "input.pdf"), 'wb') as file:
                file.write(raw_pdf)
            output = signer_setting['OUTPUT']
            cmd = [i.format(output=output) for i in signer_setting['COMMAND']]
            subprocess.run(cmd, cwd=tmp_dir, check=True, env=signer_setting['ENVIRONMENT'])  # nosec
            with open(os.path.join(tmp_dir, output), 'rb') as file:
                return file.read()
