import os
import types
from typing import Optional
from unittest.mock import ANY, _Call, call, patch, sentinel
from uuid import UUID

from django.core.exceptions import ImproperlyConfigured
from django.template.backends.django import Template as DjangoTemplate  # type: ignore[attr-defined]
from django.template.loaders.cached import Loader
from django.test import RequestFactory, TestCase, override_settings
from rest_framework.renderers import BaseRenderer
from rest_framework.request import Request
from rest_framework.response import Response
from testfixtures import Comparison, LogCapture

from django_secretary.renderers import PdfRenderer, SignedPdfRenderer, TemplateRendererMixin, TextRenderer
from django_secretary.utils.templates import Template, TemplateLoader
from django_secretary.views import TemplateViewSet

from .utils import TEMPLATES, RenderTestMixin, create_template


class TestTemplateRenderer(TemplateRendererMixin, BaseRenderer):
    """Simple renderer for tests of TemplateRendererMixin."""


class PreferredTestTemplateRenderer(TestTemplateRenderer):
    """Simple renderer with template backend preference."""
    preferred_backend = 'prefer'


@override_settings(SECRETARY_TEMPLATES=TEMPLATES)
class TemplateRendererMixinTest(RenderTestMixin, TestCase):
    def setUp(self):
        super().setUp()

        self.uuid = UUID(int=42)
        self.model = create_template(name='example.txt', content="Hello, I'm {{ name }}.", uuid=self.uuid)
        request = Request(RequestFactory().get('/dummy/'))
        self.view = TemplateViewSet(request=request, kwargs={'uuid': str(self.uuid)})
        self.renderer = TestTemplateRenderer()

    def test_get_backend(self):
        backend = self.renderer._get_backend()

        # Check secretary backend is returned.
        self.assertEqual(backend.name, 'django')
        self.assertIsInstance(backend.engine.template_loaders[0], TemplateLoader)

    def test_get_backend_none_defined(self):
        # Test error is returned when no template backend is defined.
        with override_settings(SECRETARY_TEMPLATES=[]):
            with self.assertRaisesRegex(ImproperlyConfigured, "No secretary template backend is defined."):
                self.renderer._get_backend()

    def test_get_backend_preferred(self):
        # Test preferred backend is returned when available.
        renderer = PreferredTestTemplateRenderer()
        templates = TEMPLATES.copy()
        templates.append({'BACKEND': 'django.template.backends.django.DjangoTemplates', 'NAME': 'prefer'})

        with override_settings(SECRETARY_TEMPLATES=templates):
            backend = renderer._get_backend()

        # Check secretary backend is returned.
        self.assertEqual(backend.name, 'prefer')

    def test_get_backend_preferred_unavailable(self):
        # Test preferred backend is returned when available.
        renderer = PreferredTestTemplateRenderer()

        backend = renderer._get_backend()

        # Check secretary backend is returned.
        self.assertEqual(backend.name, 'django')
        self.assertIsInstance(backend.engine.template_loaders[0], TemplateLoader)

    def test_get_template(self):
        renderer_context = {'response': Response(), 'view': self.view}

        template = self.renderer._get_template(renderer_context)

        self.assertIsInstance(template, Template)
        self.assertEqual(template.template.name, 'example.txt')
        self.assertEqual(template.model, self.model)
        # Check secretary engine is used.
        self.assertIsInstance(template.backend.engine.template_loaders[0], TemplateLoader)

    def test_get_template_error(self):
        renderer_context = {'response': Response(status=500, exception=True), 'view': self.view}

        template = self.renderer._get_template(renderer_context)

        self.assertIsInstance(template, DjangoTemplate)
        self.assertEqual(template.template.name, 'django_secretary/render_error.txt')
        # Check standard Django engine is used.
        self.assertIsInstance(template.backend.engine.template_loaders[0], Loader)

    def test_get_template_context(self):
        renderer_context = {'response': Response(), 'view': self.view}

        self.assertEqual(self.renderer._get_template_context(sentinel.data, renderer_context), sentinel.data)

    def test_get_template_context_error(self):
        renderer_context = {'response': Response(status=500, exception=True), 'view': self.view}

        self.assertEqual(self.renderer._get_template_context(sentinel.data, renderer_context),
                         {'errors': sentinel.data})


@override_settings(SECRETARY_TEMPLATES=TEMPLATES)
class TestTextRenderer(RenderTestMixin, TestCase):
    def test_render(self):
        uuid = UUID(int=42)
        create_template(name='example.txt', content="Hello, I'm {{ name }}.", uuid=uuid)
        data = {'name': 'Kryten'}
        request = Request(RequestFactory().get('/dummy/'))
        view = TemplateViewSet(request=request, kwargs={'uuid': str(uuid)})
        renderer_context = {
            'response': Response(),
            'view': view,
        }
        renderer = TextRenderer()

        result = renderer.render(data, renderer_context=renderer_context)

        self.assertEqual(result, "Hello, I'm Kryten.")
        self.assertEqual(self.template_rendered['templates'][0].name, 'example.txt')
        self.assertEqual(self.template_rendered['context'][0]['name'], 'Kryten')

    def test_render_error(self):
        renderer_context = {
            'response': Response(status=500, exception=True),
        }
        renderer = TextRenderer()
        data = {'field_42': ['Invalid value']}

        result = renderer.render(data, renderer_context=renderer_context)

        self.assertEqual(result.strip(), str(data))
        self.assertEqual(self.template_rendered['templates'][0].name, 'django_secretary/render_error.txt')
        self.assertEqual(self.template_rendered['context'][0]['errors'], data)


@override_settings(SECRETARY_TEMPLATES=TEMPLATES)
class TestPdfRenderer(RenderTestMixin, TestCase):
    def test_render(self):
        uuid = UUID(int=42)
        create_template(name='example.html', content="Hello, I'm {{ name }}.", uuid=uuid)
        data = {'name': 'Kryten'}
        request = Request(RequestFactory().get('/dummy/'))
        view = TemplateViewSet(request=request, kwargs={'uuid': str(uuid)})
        renderer_context = {
            'request': request,
            'response': Response(),
            'view': view,
        }
        renderer = PdfRenderer()

        with patch('django_secretary.renderers.weasyprint.HTML') as weasy_mock:
            weasy_mock.return_value.write_pdf.return_value = sentinel.pdf
            result = renderer.render(data, renderer_context=renderer_context)

        self.assertEqual(result, sentinel.pdf)
        self.assertEqual(self.template_rendered['templates'][0].name, 'example.html')
        self.assertEqual(self.template_rendered['context'][0]['name'], 'Kryten')
        calls = [call(string="Hello, I'm Kryten.", url_fetcher=Comparison(types.MethodType),
                      base_url='http://testserver/dummy/'),
                 call().write_pdf()]
        self.assertEqual(weasy_mock.mock_calls, calls)

    def test_render_error(self):
        renderer_context = {
            'request': Request(RequestFactory().get('/dummy/')),
            'response': Response(status=500, exception=True),
        }
        renderer = PdfRenderer()
        data = {'field_42': ['Invalid value']}

        with patch('django_secretary.renderers.weasyprint.HTML') as weasy_mock:
            weasy_mock.return_value.write_pdf.return_value = sentinel.pdf
            result = renderer.render(data, renderer_context=renderer_context)

        self.assertEqual(result, sentinel.pdf)
        self.assertEqual(self.template_rendered['templates'][0].name, 'django_secretary/render_error.txt')
        self.assertEqual(self.template_rendered['context'][0]['errors'], data)


@override_settings(SECRETARY_TEMPLATES=TEMPLATES)
class SignedPdfRendererTest(RenderTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.log_handler = LogCapture('django_secretary', propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()
        super().tearDown()

    def _test_render(self, mock_call: _Call) -> None:
        uuid = UUID(int=42)
        create_template(name='example.html', content="Hello, I'm {{ name }}.", uuid=uuid, signer='signer')
        request = Request(RequestFactory().get('/dummy/'))
        view = TemplateViewSet(request=request, kwargs={'uuid': str(uuid)})
        renderer_context = {
            'request': request,
            'response': Response(),
            'view': view,
        }
        renderer = SignedPdfRenderer()
        content = b'Gazpacho!'

        with patch('subprocess.run') as run_mock:
            with patch('django_secretary.renderers.weasyprint.HTML') as weasy_mock:
                weasy_mock.return_value.write_pdf.return_value = content
                result = renderer.render({'name': 'Kryten'}, renderer_context=renderer_context)

        self.assertEqual(result, content)
        self.assertEqual(run_mock.mock_calls, [mock_call])
        # Check temporary directory was deleted.
        self.assertFalse(os.path.exists(run_mock.mock_calls[0][2]['cwd']))

    def test_render(self):
        signer = {'COMMAND': ['hollister', 'input.pdf', '{output}'],
                  # Set the output name to the same as input.
                  'OUTPUT': 'input.pdf'}
        with override_settings(SECRETARY_SIGNERS={'signer': signer}):
            mock_call = call(['hollister', 'input.pdf', 'input.pdf'], cwd=ANY, check=True, env=None)
            self._test_render(mock_call)

    def test_render_environment(self):
        signer = {'COMMAND': ['hollister', 'input.pdf', '{output}'],
                  # Set the output name to the same as input.
                  'OUTPUT': 'input.pdf',
                  'ENVIRONMENT': {'FOO': 'BAR'}}
        with override_settings(SECRETARY_SIGNERS={'signer': signer}):
            mock_call = call(['hollister', 'input.pdf', 'input.pdf'], cwd=ANY, check=True, env={'FOO': 'BAR'})
            self._test_render(mock_call)

    def _test_render_unsigned(self, signer: Optional[str]) -> None:
        uuid = UUID(int=42)
        create_template(name='example.html', content="Hello, I'm {{ name }}.", uuid=uuid, signer=signer)
        request = Request(RequestFactory().get('/dummy/'))
        view = TemplateViewSet(request=request, kwargs={'uuid': str(uuid)})
        renderer_context = {
            'request': request,
            'response': Response(),
            'view': view,
        }
        renderer = SignedPdfRenderer()

        with patch('subprocess.run') as run_mock:
            with patch('django_secretary.renderers.weasyprint.HTML') as weasy_mock:
                weasy_mock.return_value.write_pdf.return_value = sentinel.pdf
                result = renderer.render({'name': 'Kryten'}, renderer_context=renderer_context)

        self.assertEqual(result, sentinel.pdf)
        self.assertEqual(run_mock.mock_calls, [])

    def test_render_unsigned_no_signer(self):
        with override_settings(SECRETARY_SIGNERS={}):
            self._test_render_unsigned(None)

    def test_render_unsigned_unknown_signer(self):
        with override_settings(SECRETARY_SIGNERS={}):
            self._test_render_unsigned('unknown')

    def test_render_error(self):
        renderer_context = {
            'request': Request(RequestFactory().get('/dummy/')),
            'response': Response(status=500, exception=True),
        }
        renderer = SignedPdfRenderer()

        with patch('subprocess.run') as run_mock:
            with patch('django_secretary.renderers.weasyprint.HTML') as weasy_mock:
                weasy_mock.return_value.write_pdf.return_value = sentinel.pdf
                result = renderer.render({'name': 'Kryten'}, renderer_context=renderer_context)

        self.assertEqual(result, sentinel.pdf)
        self.assertEqual(run_mock.mock_calls, [])
